@@ -1,3 +1,16 @@
+/*
+ * LibHTP (http://www.libhtp.org)
+ * Copyright 2009,2010 Ivan Ristic <ivanr@webkreator.com>
+ *
+ * LibHTP is an open source product, released under terms of the General Public Licence
+ * version 2 (GPLv2). Please refer to the file LICENSE, which contains the complete text
+ * of the license.
+ *
+ * In addition, there is a special exception that allows LibHTP to be freely
+ * used with any OSI-approved open source licence. Please refer to the file
+ * LIBHTP_LICENSING_EXCEPTION for the full text of the exception.
+ *
+ */
 
 #ifndef _HTP_H
 #define	_HTP_H
@@ -28,10 +41,13 @@ typedef struct htp_urldecoder_t htp_urldecoder_t;
 
 // -- Defines -------------------------------------------------------------------------------------
 
+#define HTP_BASE_VERSION_TEXT	"0.2.6"
+
 #define HTP_ERROR              -1
 #define HTP_OK                  0
 #define HTP_DATA                1
-#define HTP_DECLINED            2
+#define HTP_DATA_OTHER          2
+#define HTP_DECLINED            3
 
 #define PROTOCOL_UNKNOWN        -1
 #define HTTP_0_9                9
@@ -158,6 +174,8 @@ typedef struct htp_urldecoder_t htp_urldecoder_t;
 #define STREAM_STATE_OPEN           1
 #define STREAM_STATE_CLOSED         2
 #define STREAM_STATE_ERROR          3
+#define STREAM_STATE_TUNNEL         4
+#define STREAM_STATE_DATA_OTHER     5
 #define STREAM_STATE_DATA           9
 
 #define URL_DECODER_PRESERVE_PERCENT            0
@@ -355,6 +373,9 @@ struct htp_cfg_t {
     /** The best-fit map to use to decode %u-encoded characters. */
     unsigned char *path_u_bestfit_map;
 
+    /** Whether to generate the request_uri_normalized field. */
+    int generate_request_uri_normalized;
+
 
     // Hooks
 
@@ -366,6 +387,9 @@ struct htp_cfg_t {
     /** Request line hook, invoked after a request line has been parsed. */
     htp_hook_t *hook_request_line;
 
+    /** Request URI normalization hook, for overriding default normalization of URI. */
+    htp_hook_t *hook_request_uri_normalize;
+
     /** Request headers hook, invoked after all request headers are seen. */
     htp_hook_t *hook_request_headers;
 
@@ -736,6 +760,13 @@ struct htp_tx_t {
     /** Request URI, raw, as given to us on the request line. */
     bstr *request_uri;
 
+    /**
+     * Normalized request URI as a single string. The availability of this
+     * field depends on configuration. Use htp_config_set_generate_request_uri_normalized()
+     * to ask for the field to be generated.
+     */
+    bstr *request_uri_normalized;
+
     /** Request protocol, as text. */
     bstr *request_protocol;
 
@@ -797,6 +828,17 @@ struct htp_tx_t {
     /** Parsed request headers. */
     table_t *request_headers;
 
+    /** Contains raw request headers. This field is generated on demand, use
+     *  htp_tx_get_request_headers_raw() to get it.
+     */
+    bstr *request_headers_raw;
+
+    /** How many request header lines have been included in the raw
+     *  buffer (above).
+     */
+    size_t request_headers_raw_lines;
+
+
     /** Request transfer coding: IDENTITY or CHUNKED. Only available on requests that have bodies. */
     int request_transfer_coding;
 
@@ -928,6 +970,7 @@ htp_cfg_t *htp_config_create();
 
 void htp_config_register_transaction_start(htp_cfg_t *cfg, int (*callback_fn)(htp_connp_t *));
 void htp_config_register_request_line(htp_cfg_t *cfg, int (*callback_fn)(htp_connp_t *));
+void htp_config_register_request_uri_normalize(htp_cfg_t *cfg, int (*callback_fn)(htp_connp_t *));
 void htp_config_register_request_headers(htp_cfg_t *cfg, int (*callback_fn)(htp_connp_t *));
 void htp_config_register_request_body_data(htp_cfg_t *cfg, int (*callback_fn)(htp_tx_data_t *));
 void htp_config_register_request_trailer(htp_cfg_t *cfg, int (*callback_fn)(htp_connp_t *));
@@ -958,6 +1001,8 @@ void htp_config_set_path_nul_raw_handling(htp_cfg_t *cfg, int nul_raw_handling);
 void htp_config_set_path_replacement_char(htp_cfg_t *cfg, int replacement_char);
 void htp_config_set_path_unicode_mapping(htp_cfg_t *cfg, int unicode_mapping);
 
+void htp_config_set_generate_request_uri_normalized(htp_cfg_t *cfg, int generate);
+
 
 htp_connp_t *htp_connp_create(htp_cfg_t *cfg);
 htp_connp_t *htp_connp_create_copycfg(htp_cfg_t *cfg);
@@ -973,10 +1018,10 @@ htp_conn_t *htp_conn_create();
        void htp_conn_destroy(htp_conn_t *conn);
         int htp_conn_remove_tx(htp_conn_t *conn, htp_tx_t *tx);
 
-int htp_connp_req_data(htp_connp_t *connp, htp_time_t timestamp, unsigned char *data, size_t len);
-// int htp_connp_req_data_missing(htp_connp_t *connp, htp_time_t timestamp, size_t len);
-int htp_connp_res_data(htp_connp_t *connp, htp_time_t timestamp, unsigned char *data, size_t len);
-// int htp_connp_res_data_missing(htp_connp_t *connp, htp_time_t timestamp, size_t len);
+   int htp_connp_req_data(htp_connp_t *connp, htp_time_t timestamp, unsigned char *data, size_t len);
+size_t htp_connp_req_data_consumed(htp_connp_t *connp);
+   int htp_connp_res_data(htp_connp_t *connp, htp_time_t timestamp, unsigned char *data, size_t len);
+size_t htp_connp_res_data_consumed(htp_connp_t *connp);
 
       void htp_connp_clear_error(htp_connp_t *connp);
 htp_log_t *htp_connp_get_last_error(htp_connp_t *connp);
@@ -993,6 +1038,8 @@ htp_tx_t *htp_tx_create(htp_cfg_t *cfg, int is_cfg_shared, htp_conn_t *conn);
      void htp_tx_set_user_data(htp_tx_t *tx, void *user_data);
     void *htp_tx_get_user_data(htp_tx_t *tx);
 
+    bstr *htp_tx_get_request_uri_normalized(htp_tx_t *tx);
+
 // Parsing functions
 
 int htp_parse_request_line_generic(htp_connp_t *connp);
@@ -1018,6 +1065,9 @@ int htp_connp_REQ_BODY_CHUNKED_LENGTH(htp_connp_t *connp);
 int htp_connp_REQ_BODY_CHUNKED_DATA(htp_connp_t *connp);
 int htp_connp_REQ_BODY_CHUNKED_DATA_END(htp_connp_t *connp);
 
+int htp_connp_REQ_CONNECT_CHECK(htp_connp_t *connp);
+int htp_connp_REQ_CONNECT_WAIT_RESPONSE(htp_connp_t *connp);
+
 int htp_connp_RES_IDLE(htp_connp_t *connp);
 int htp_connp_RES_LINE(htp_connp_t *connp);
 int htp_connp_RES_HEADERS(htp_connp_t *connp);
@@ -1076,6 +1126,10 @@ char *htp_connp_in_state_as_string(htp_connp_t *connp);
 char *htp_connp_out_state_as_string(htp_connp_t *connp);
 char *htp_tx_progress_as_string(htp_tx_t *tx);
 
+bstr *htp_unparse_uri_noencode(htp_uri_t *uri);
+
+bstr *htp_tx_get_request_headers_raw(htp_tx_t *tx);
+
 #endif	/* _HTP_H */
 
 
