@@ -1,12 +1,25 @@
+/*
+ * LibHTP (http://www.libhtp.org)
+ * Copyright 2009,2010 Ivan Ristic <ivanr@webkreator.com>
+ *
+ * LibHTP is an open source product, released under terms of the General Public Licence
+ * version 2 (GPLv2). Please refer to the file LICENSE, which contains the complete text
+ * of the license.
+ *
+ * In addition, there is a special exception that allows LibHTP to be freely
+ * used with any OSI-approved open source licence. Please refer to the file
+ * LIBHTP_LICENSING_EXCEPTION for the full text of the exception.
+ *
+ */
 
 #include <stdlib.h>
-
 #include "htp.h"
 
 /**
  * Invoked whenever decompressed response body data becomes available.
  *
  * @param d
+ * @return HTP_OK on state change, HTP_ERROR on error.
  */
 static int htp_connp_RES_BODY_DECOMPRESSOR_CALLBACK(htp_tx_data_t *d) {
     // Invoke all callbacks
@@ -242,7 +255,21 @@ int htp_connp_RES_BODY_IDENTITY(htp_connp_t *connp) {
  * @returns HTP_OK on state change, HTTP_ERROR on error, or HTP_DATA when more data is needed.
  */
 int htp_connp_RES_BODY_DETERMINE(htp_connp_t *connp) {
-    // First check for an interim "100 Continue"
+    // If the request uses the CONNECT method, then not only are we
+    // to assume there's no body, but we need to ignore all
+    // subsequent data in the stream.
+    if ((connp->out_tx->request_method_number == M_CONNECT)
+        &&(connp->out_tx->response_status_number >= 200)
+        &&(connp->out_tx->response_status_number <= 299))
+    {
+        connp->out_status = STREAM_STATE_TUNNEL;
+        connp->out_state = htp_connp_RES_IDLE;
+        connp->out_tx->progress = TX_PROGRESS_DONE;
+
+        return HTP_OK;
+    }
+
+    // Check for an interim "100 Continue"
     // response. Ignore it if found, and revert back to RES_FIRST_LINE.
     if (connp->out_tx->response_status_number == 100) {
         if (connp->out_tx->seen_100continue != 0) {
@@ -414,6 +441,10 @@ int htp_connp_RES_HEADERS(htp_connp_t *connp) {
 
         // Have we reached the end of the line?
         if (connp->out_next_byte == LF) {
+            #ifdef HTP_DEBUG
+            fprint_raw_data(stderr, __FUNCTION__, connp->out_line, connp->out_line_len);
+            #endif
+
             // Should we terminate headers?
             if (htp_connp_is_line_terminator(connp, connp->out_line, connp->out_line_len)) {
                 // Terminator line
@@ -432,9 +463,7 @@ int htp_connp_RES_HEADERS(htp_connp_t *connp) {
                 // Cleanup
                 free(connp->out_header_line);
                 connp->out_line_len = 0;
-                connp->out_header_line = NULL;
-                connp->out_header_line_index = -1;
-                connp->out_header_line_counter = 0;
+                connp->out_header_line = NULL;                
 
                 // We've seen all response headers
                 if (connp->out_tx->progress == TX_PROGRESS_RES_HEADERS) {
@@ -488,6 +517,9 @@ int htp_connp_RES_HEADERS(htp_connp_t *connp) {
 
             // Add the raw header line to the list
             connp->out_header_line->line = bstr_memdup((char *) connp->out_line, connp->out_line_len);
+            if (connp->out_header_line->line == NULL) {
+                return HTP_ERROR;
+            }
             list_add(connp->out_tx->response_header_lines, connp->out_header_line);
             connp->out_header_line = NULL;
 
@@ -515,6 +547,10 @@ int htp_connp_RES_LINE(htp_connp_t *connp) {
 
         // Have we reached the end of the line?
         if (connp->out_next_byte == LF) {
+            #ifdef HTP_DEBUG
+            fprint_raw_data(stderr, __FUNCTION__, connp->out_line, connp->out_line_len);
+            #endif
+
             // Is this a line that should be ignored?
             if (htp_connp_is_line_ignorable(connp, connp->out_line, connp->out_line_len)) {
                 // We have an empty/whitespace line, which we'll note, ignore and move on
@@ -551,6 +587,9 @@ int htp_connp_RES_LINE(htp_connp_t *connp) {
             }
 
             connp->out_tx->response_line = bstr_memdup((char *) connp->out_line, connp->out_line_len);
+            if (connp->out_tx->response_line == NULL) {
+                return HTP_ERROR;
+            }
 
             // Parse response line
             if (connp->cfg->parse_response_line(connp) != HTP_OK) {
@@ -578,6 +617,10 @@ int htp_connp_RES_LINE(htp_connp_t *connp) {
     }
 }
 
+size_t htp_connp_res_data_consumed(htp_connp_t *connp) {
+    return connp->out_current_offset;
+}
+
 /**
  * The response idle state will initialize response processing, as well as
  * finalize each transactions after we are done with it.
@@ -606,6 +649,15 @@ int htp_connp_RES_IDLE(htp_connp_t * connp) {
             return HTP_ERROR;
         }
 
+        // Check if the inbound parser is waiting on us. If it is that means that
+        // there might be request data that the inbound parser hasn't consumed yet.
+        // If we don't stop parsing we might encounter a response without a
+        // request.
+        if ((connp->in_status == STREAM_STATE_DATA_OTHER) && (connp->in_tx == connp->out_tx)) {
+            connp->out_tx = NULL;
+            return HTP_DATA_OTHER;
+        }
+
         // Start afresh
         connp->out_tx = NULL;
     }
@@ -662,17 +714,32 @@ int htp_connp_RES_IDLE(htp_connp_t * connp) {
  * @return HTP_OK on state change, HTTP_ERROR on error, or HTP_DATA when more data is needed
  */
 int htp_connp_res_data(htp_connp_t *connp, htp_time_t timestamp, unsigned char *data, size_t len) {
-#ifdef HTP_DEBUG
-    fprintf(stderr, "htp_connp_req_data(connp->out_status %x)\n", connp->out_status);
+    #ifdef HTP_DEBUG
+    fprintf(stderr, "htp_connp_res_data(connp->out_status %x)\n", connp->out_status);
     fprint_raw_data(stderr, __FUNCTION__, data, len);
-#endif
+    #endif
 
     // Return if the connection has had a fatal error
-    if (connp->out_status != STREAM_STATE_OPEN) {
-        // We allow calls that allow the parser to finalize their work
-        if (!((connp->out_status == STREAM_STATE_CLOSED) && (len == 0))) {
-            return STREAM_STATE_ERROR;
-        }
+    if (connp->out_status == STREAM_STATE_ERROR) {
+        htp_log(connp, HTP_LOG_MARK, HTP_LOG_ERROR, 0, "Outbound parser is in STREAM_STATE_ERROR");
+
+        #ifdef HTP_DEBUG
+        fprintf(stderr, "htp_connp_res_data: returning STREAM_STATE_DATA (previous error)\n");
+        #endif
+        return STREAM_STATE_ERROR;
+    }
+
+    // If the length of the supplied data chunk is zero, proceed
+    // only if the stream has been closed. We do not allow zero-sized
+    // chunks in the API, but we use it internally to force the parsers
+    // to finalize parsing.
+    if ((len == 0) && (connp->out_status != STREAM_STATE_CLOSED)) {
+        htp_log(connp, HTP_LOG_MARK, HTP_LOG_ERROR, 0, "Zero-length data chunks are not allowed");
+
+        #ifdef HTP_DEBUG
+        fprintf(stderr, "htp_connp_res_data: returning STREAM_STATE_DATA (zero-length chunk)\n");
+        #endif
+        return STREAM_STATE_ERROR;
     }
 
     // Store the current chunk information
@@ -683,16 +750,25 @@ int htp_connp_res_data(htp_connp_t *connp, htp_time_t timestamp, unsigned char *
     connp->conn->out_data_counter += len;
     connp->conn->out_packet_counter++;
 
+    // Return without processing any data if the stream is in tunneling
+    // mode (which it would be after an initial CONNECT transaction.
+    if (connp->out_status == STREAM_STATE_TUNNEL) {
+        #ifdef HTP_DEBUG
+        fprintf(stderr, "htp_connp_res_data: returning STREAM_STATE_DATA (tunnel)\n");
+        #endif
+        return STREAM_STATE_DATA;
+    }
+
     // Invoke a processor, in a loop, until an error
     // occurs or until we run out of data. Many processors
     // will process a request, each pointing to the next
     // processor that needs to run.
     for (;;) {
-#ifdef HTP_DEBUG
+        #ifdef HTP_DEBUG
         fprintf(stderr, "htp_connp_res_data: out state=%s, progress=%s\n",
             htp_connp_out_state_as_string(connp),
             htp_tx_progress_as_string(connp->out_tx));
-#endif
+        #endif
         // Return if there's been an error
         // or if we've run out of data. We are relying
         // on processors to add error messages, so we'll
@@ -704,6 +780,25 @@ int htp_connp_res_data(htp_connp_t *connp, htp_time_t timestamp, unsigned char *
                 return STREAM_STATE_DATA;
             }
 
+            // Check for suspended parsing
+            if (rc == HTP_DATA_OTHER) {
+                // We might have actually consumed the entire data chunk?
+                if (connp->out_current_offset >= connp->out_current_len) {
+                    // Do not send STREAM_DATE_DATA_OTHER if we've
+                    // consumed the entire chunk
+                    #ifdef HTP_DEBUG
+                    fprintf(stderr, "htp_connp_res_data: returning STREAM_STATE_DATA (suspended parsing)\n");
+                    #endif
+                    return STREAM_STATE_DATA;
+                } else {
+                    // Partial chunk consumption
+                    #ifdef HTP_DEBUG
+                    fprintf(stderr, "htp_connp_req_data: returning STREAM_STATE_DATA_OTHER\n");
+                    #endif
+                    return STREAM_STATE_DATA_OTHER;
+                }
+            }
+
             // Remember that we've had an error. Errors are
             // not possible to recover from.
             connp->out_status = STREAM_STATE_ERROR;
