@@ -1,3 +1,16 @@
+/*
+ * LibHTP (http://www.libhtp.org)
+ * Copyright 2009,2010 Ivan Ristic <ivanr@webkreator.com>
+ *
+ * LibHTP is an open source product, released under terms of the General Public Licence
+ * version 2 (GPLv2). Please refer to the file LICENSE, which contains the complete text
+ * of the license.
+ *
+ * In addition, there is a special exception that allows LibHTP to be freely
+ * used with any OSI-approved open source licence. Please refer to the file
+ * LIBHTP_LICENSING_EXCEPTION for the full text of the exception.
+ *
+ */
 
 #include "htp.h"
 #include "utf8_decoder.h"
@@ -293,6 +306,8 @@ void htp_log(htp_connp_t *connp, const char *file, int line, int level, int code
     va_start(args, fmt);
 
     int r = vsnprintf(buf, 1023, fmt, args);
+    
+    va_end(args);
 
     if (r < 0) {
         // TODO Will vsnprintf ever return an error?
@@ -322,9 +337,7 @@ void htp_log(htp_connp_t *connp, const char *file, int line, int level, int code
         connp->last_error = log;
     }
 
-    hook_run_all(connp->cfg->hook_log, log);
-
-    va_end(args);
+    hook_run_all(connp->cfg->hook_log, log);   
 }
 
 /**
@@ -422,7 +435,7 @@ int htp_parse_authority(htp_connp_t *connp, bstr *authority, htp_uri_t **uri) {
         }
     }
 
-    return 1;
+    return HTP_OK;
 }
 
 /**
@@ -479,63 +492,63 @@ int htp_parse_uri(bstr *input, htp_uri_t **uri) {
     // One, three or more slash characters, and it's a path. We, however,
     // only attempt to parse authority if we've seen a scheme.
     if ((*uri)->scheme != NULL)
-    if ((pos + 2 < len) && (data[pos] == '/') && (data[pos + 1] == '/') && (data[pos + 2] != '/')) {
-        // Parse authority
+        if ((pos + 2 < len) && (data[pos] == '/') && (data[pos + 1] == '/') && (data[pos + 2] != '/')) {
+            // Parse authority
 
-        // Go over the two slash characters
-        start = pos = pos + 2;
+            // Go over the two slash characters
+            start = pos = pos + 2;
 
-        // Authority ends with a question mark, forward slash or hash
-        while ((pos < len) && (data[pos] != '?') && (data[pos] != '/') && (data[pos] != '#')) pos++;
+            // Authority ends with a question mark, forward slash or hash
+            while ((pos < len) && (data[pos] != '?') && (data[pos] != '/') && (data[pos] != '#')) pos++;
 
-        char *hostname_start;
-        size_t hostname_len;
+            char *hostname_start;
+            size_t hostname_len;
 
-        // Are the credentials included in the authority?
-        char *m = memchr(data + start, '@', pos - start);
-        if (m != NULL) {
-            // Credentials present
-            char *credentials_start = data + start;
-            size_t credentials_len = m - data - start;
-
-            // Figure out just the hostname part
-            hostname_start = data + start + credentials_len + 1;
-            hostname_len = pos - start - credentials_len - 1;
-
-            // Extract the username and the password
-            m = memchr(credentials_start, ':', credentials_len);
+            // Are the credentials included in the authority?
+            char *m = memchr(data + start, '@', pos - start);
             if (m != NULL) {
-                // Username and password
-                (*uri)->username = bstr_memdup(credentials_start, m - credentials_start);
-                (*uri)->password = bstr_memdup(m + 1, credentials_len - (m - credentials_start) - 1);
+                // Credentials present
+                char *credentials_start = data + start;
+                size_t credentials_len = m - data - start;
+
+                // Figure out just the hostname part
+                hostname_start = data + start + credentials_len + 1;
+                hostname_len = pos - start - credentials_len - 1;
+
+                // Extract the username and the password
+                m = memchr(credentials_start, ':', credentials_len);
+                if (m != NULL) {
+                    // Username and password
+                    (*uri)->username = bstr_memdup(credentials_start, m - credentials_start);
+                    (*uri)->password = bstr_memdup(m + 1, credentials_len - (m - credentials_start) - 1);
+                } else {
+                    // Username alone
+                    (*uri)->username = bstr_memdup(credentials_start, credentials_len);
+                }
             } else {
-                // Username alone
-                (*uri)->username = bstr_memdup(credentials_start, credentials_len);
+                // No credentials
+                hostname_start = data + start;
+                hostname_len = pos - start;
             }
-        } else {
-            // No credentials
-            hostname_start = data + start;
-            hostname_len = pos - start;
+
+            // Still parsing authority; is there a port provided?
+            m = memchr(hostname_start, ':', hostname_len);
+            if (m != NULL) {
+                size_t port_len = hostname_len - (m - hostname_start) - 1;
+                hostname_len = hostname_len - port_len - 1;
+
+                // Port string
+                (*uri)->port = bstr_memdup(m + 1, port_len);
+
+                // We deliberately don't want to try to convert the port
+                // string as a number. That will be done later, during
+                // the normalization and validation process.
+            }
+
+            // Hostname
+            (*uri)->hostname = bstr_memdup(hostname_start, hostname_len);
         }
 
-        // Still parsing authority; is there a port provided?
-        m = memchr(hostname_start, ':', hostname_len);
-        if (m != NULL) {
-            size_t port_len = hostname_len - (m - hostname_start) - 1;
-            hostname_len = hostname_len - port_len - 1;
-
-            // Port string
-            (*uri)->port = bstr_memdup(m + 1, port_len);
-
-            // We deliberately don't want to try to convert the port
-            // string as a number. That will be done later, during
-            // the normalization and validation process.
-        }
-
-        // Hostname
-        (*uri)->hostname = bstr_memdup(hostname_start, hostname_len);
-    }
-
     // Path
     start = pos;
 
@@ -648,7 +661,7 @@ void htp_utf8_decode_path_inplace(htp_cfg_t *cfg, htp_tx_t *tx, bstr *path) {
     size_t rpos = 0;
     size_t wpos = 0;
     size_t charpos = 0;
-    uint32_t codepoint;
+    uint32_t codepoint = 0;
     uint32_t state = UTF8_ACCEPT;
     uint32_t counter = 0;
     uint8_t seen_valid = 0;
@@ -677,6 +690,7 @@ void htp_utf8_decode_path_inplace(htp_cfg_t *cfg, htp_tx_t *tx, bstr *path) {
                             if (codepoint < 0x800) {
                                 tx->flags |= HTP_PATH_UTF8_OVERLONG;
                             }
+                            break;
                         case 4:
                             if (codepoint < 0x10000) {
                                 tx->flags |= HTP_PATH_UTF8_OVERLONG;
@@ -720,12 +734,8 @@ void htp_utf8_decode_path_inplace(htp_cfg_t *cfg, htp_tx_t *tx, bstr *path) {
                     data[wpos++] = data[charpos++];
                 }
 
-                // If this is the first invalid byte we will
-                // want to skip over it. Otherwise we will want
-                // to attempt to interpret it.
-                if (counter == 1) {
-                    rpos++;
-                }
+                // Advance over the consumed byte
+                rpos++;
 
                 // Prepare for the next character
                 counter = 0;
@@ -741,7 +751,7 @@ void htp_utf8_decode_path_inplace(htp_cfg_t *cfg, htp_tx_t *tx, bstr *path) {
     }
 
     // Did the input stream seem like a valid UTF-8 string?
-    if ((seen_valid) && (!(tx->flags | HTP_PATH_UTF8_INVALID))) {
+    if ((seen_valid) && (!(tx->flags & HTP_PATH_UTF8_INVALID))) {
         tx->flags |= HTP_PATH_UTF8_VALID;
     }
 
@@ -761,8 +771,7 @@ void htp_utf8_validate_path(htp_tx_t *tx, bstr *path) {
     unsigned char *data = (unsigned char *) bstr_ptr(path);
     size_t len = bstr_len(path);
     size_t rpos = 0;
-    size_t charpos = 0;
-    uint32_t codepoint;
+    uint32_t codepoint = 0;
     uint32_t state = UTF8_ACCEPT;
     uint32_t counter = 0;
     uint8_t seen_valid = 0;
@@ -790,6 +799,7 @@ void htp_utf8_validate_path(htp_tx_t *tx, bstr *path) {
                             if (codepoint < 0x800) {
                                 tx->flags |= HTP_PATH_UTF8_OVERLONG;
                             }
+                            break;
                         case 4:
                             if (codepoint < 0x10000) {
                                 tx->flags |= HTP_PATH_UTF8_OVERLONG;
@@ -808,7 +818,6 @@ void htp_utf8_validate_path(htp_tx_t *tx, bstr *path) {
 
                 // Prepare for the next character
                 counter = 0;
-                charpos = rpos;
 
                 break;
 
@@ -820,16 +829,11 @@ void htp_utf8_validate_path(htp_tx_t *tx, bstr *path) {
                 // we want to ignore invalid characters
                 state = UTF8_ACCEPT;
 
-                // If this is the first invalid byte we will
-                // want to skip over it. Otherwise we will want
-                // to attempt to interpret it.
-                if (counter == 1) {
-                    rpos++;
-                }
+                // Advance over the consumed byte
+                rpos++;
 
                 // Prepare for the next character
                 counter = 0;
-                charpos = rpos;
 
                 break;
 
@@ -841,7 +845,7 @@ void htp_utf8_validate_path(htp_tx_t *tx, bstr *path) {
     }
 
     // Did the input stream seem like a valid UTF-8 string?
-    if ((seen_valid) && (!(tx->flags | HTP_PATH_UTF8_INVALID))) {
+    if ((seen_valid) && (!(tx->flags & HTP_PATH_UTF8_INVALID))) {
         tx->flags |= HTP_PATH_UTF8_VALID;
     }
 }
@@ -917,6 +921,9 @@ int decode_u_encoding(htp_cfg_t *cfg, htp_tx_t *tx, unsigned char *data) {
  */
 int htp_decode_path_inplace(htp_cfg_t *cfg, htp_tx_t *tx, bstr *path) {
     unsigned char *data = (unsigned char *) bstr_ptr(path);
+    if (data == NULL) {
+        return -1;
+    }
     size_t len = bstr_len(path);
 
     // TODO I don't like this function. It's too complex.
@@ -1241,6 +1248,8 @@ int htp_normalize_parsed_uri(htp_connp_t *connp, htp_uri_t *incomplete, htp_uri_
 
     // Query
     if (incomplete->query != NULL) {
+        // We cannot URL-decode the query string here; it needs to be
+        // parsed into individual key-value pairs first.
         normalized->query = bstr_strdup(incomplete->query);
     }
 
@@ -1342,8 +1351,8 @@ void htp_uriencoding_normalize_inplace(bstr *s) {
     unsigned char *data = (unsigned char *) bstr_ptr(s);
     size_t len = bstr_len(s);
 
-    size_t rpos = len;
-    size_t wpos = rpos;
+    size_t rpos = 0;
+    size_t wpos = 0;
 
     while (rpos < len) {
         if (data[rpos] == '%') {
@@ -1384,7 +1393,7 @@ void htp_uriencoding_normalize_inplace(bstr *s) {
         }
     }
 
-    bstr_len_adjust(s, wpos - 1);
+    bstr_len_adjust(s, wpos);
 }
 
 #if 0
@@ -1646,6 +1655,9 @@ char *htp_connp_in_state_as_string(htp_connp_t *connp) {
     if (connp->in_state == htp_connp_REQ_BODY_CHUNKED_DATA) return "REQ_BODY_CHUNKED_DATA";
     if (connp->in_state == htp_connp_REQ_BODY_CHUNKED_DATA_END) return "REQ_BODY_CHUNKED_DATA_END";
 
+    if (connp->in_state == htp_connp_REQ_CONNECT_CHECK) return "htp_connp_REQ_CONNECT_CHECK";
+    if (connp->in_state == htp_connp_REQ_CONNECT_WAIT_RESPONSE) return "htp_connp_REQ_CONNECT_WAIT_RESPONSE";
+
     return "UNKNOWN";
 }
 
@@ -1656,7 +1668,7 @@ char *htp_connp_out_state_as_string(htp_connp_t *connp) {
     if (connp == NULL) return "NULL";
 
     if (connp->out_state == htp_connp_RES_IDLE) return "RES_IDLE";
-    if (connp->out_state == htp_connp_RES_LINE) return "";
+    if (connp->out_state == htp_connp_RES_LINE) return "RES_LINE";
     if (connp->out_state == htp_connp_RES_HEADERS) return "RES_HEADERS";
     if (connp->out_state == htp_connp_RES_BODY_DETERMINE) return "RES_BODY_DETERMINE";
     if (connp->out_state == htp_connp_RES_BODY_IDENTITY) return "RES_BODY_IDENTITY";
@@ -1700,3 +1712,169 @@ char *htp_tx_progress_as_string(htp_tx_t *tx) {
 
     return "UNKOWN";
 }
+
+bstr *htp_unparse_uri_noencode(htp_uri_t *uri) {
+    if (uri == NULL) {
+        return NULL;
+    }   
+
+    // On the first pass determine the length of the final string
+    size_t len = 0;
+
+    if (uri->scheme != NULL) {
+        len += bstr_len(uri->scheme);
+        len += 3; // "://"
+    }
+
+    if ((uri->username != NULL) || (uri->password != NULL)) {
+        if (uri->username != NULL) {
+            len += bstr_len(uri->username);
+        }
+
+        len += 1; // ":"
+
+        if (uri->password != NULL) {
+            len += bstr_len(uri->password);
+        }
+
+        len += 1; // "@"
+    }   
+
+    if (uri->hostname != NULL) {
+        len += bstr_len(uri->hostname);
+    }   
+
+    if (uri->port != NULL) {
+        len += 1; // ":"
+        len += bstr_len(uri->port);
+    }
+
+    if (uri->path != NULL) {
+        len += bstr_len(uri->path);
+    }
+
+    if (uri->query != NULL) {
+        len += 1; // "?"
+        len += bstr_len(uri->query);
+    }
+
+    if (uri->fragment != NULL) {
+        len += 1; // "#"
+        len += bstr_len(uri->fragment);
+    }    
+
+    // On the second pass construct the string
+    bstr *r = bstr_alloc(len);
+    if (r == NULL) {
+        return NULL;
+    }   
+
+    if (uri->scheme != NULL) {
+        bstr_add_str_noex(r, uri->scheme);
+        bstr_add_cstr_noex(r, "://");
+    }   
+
+    if ((uri->username != NULL) || (uri->password != NULL)) {
+        if (uri->username != NULL) {
+            bstr_add_str_noex(r, uri->username);
+        }
+
+        bstr_add_cstr(r, ":");
+
+        if (uri->password != NULL) {
+            bstr_add_str_noex(r, uri->password);
+        }
+
+        bstr_add_cstr_noex(r, "@");
+    }   
+
+    if (uri->hostname != NULL) {
+        bstr_add_str_noex(r, uri->hostname);
+    }   
+
+    if (uri->port != NULL) {
+        bstr_add_cstr(r, ":");
+        bstr_add_str_noex(r, uri->port);
+    }   
+
+    if (uri->path != NULL) {
+        bstr_add_str_noex(r, uri->path);
+    }  
+
+    if (uri->query != NULL) {
+        bstr *query = bstr_strdup(uri->query);
+        htp_uriencoding_normalize_inplace(query);
+        bstr_add_cstr_noex(r, "?");
+        bstr_add_str_noex(r, query);
+        bstr_free(query);
+    }     
+
+    if (uri->fragment != NULL) {
+        bstr_add_cstr_noex(r, "#");
+        bstr_add_str_noex(r, uri->fragment);
+    }
+   
+    return r;
+}
+
+/**
+ * Construct a bstr that contains the raw request headers.
+ *
+ * @param tx
+ * @return
+ */
+bstr *htp_tx_generate_request_headers_raw(htp_tx_t *tx) {
+    bstr *request_headers_raw = NULL;
+    size_t i, len = 0;
+
+    for (i = 0; i < list_size(tx->request_header_lines); i++) {
+        htp_header_line_t *hl = list_get(tx->request_header_lines, i);
+        len += bstr_len(hl->line);
+    }
+
+    request_headers_raw = bstr_alloc(len);
+    if (request_headers_raw == NULL) {
+        htp_log(tx->connp, HTP_LOG_MARK, HTP_LOG_ERROR, 0, "Failed to allocate bstring of %d bytes", len);
+        return NULL;
+    }
+
+    for (i = 0; i < list_size(tx->request_header_lines); i++) {
+        htp_header_line_t *hl = list_get(tx->request_header_lines, i);
+        bstr_add_str_noex(request_headers_raw, hl->line);
+    }
+
+    return request_headers_raw;
+}
+
+/**
+ * Get a bstr that contains the raw request headers. This method will always
+ * return an up-to-date buffer, containing the last known headers. Thus, if
+ * it is called once after REQUEST_HEADERS phase it will return one buffer, but
+ * it may return a different buffer if called after REQUEST_TRAILERS phase (but
+ * only if the request actually contains trailer headers). Do not retain the
+ * bstr pointer, as the buffer may change. If there are no changes to the
+ * request header structure, only one buffer will be contstructed and used. (Multiple
+ * invocations of this method will not cause multiple buffers to be created.)
+ *
+ * @param tx
+ * @return
+ */
+bstr *htp_tx_get_request_headers_raw(htp_tx_t *tx) {
+    // Check that we are not called too early
+    if (tx->progress < TX_PROGRESS_REQ_HEADERS) return NULL;
+
+    if (tx->request_headers_raw == NULL) {
+        tx->request_headers_raw = htp_tx_generate_request_headers_raw(tx);
+        tx->request_headers_raw_lines = list_size(tx->request_header_lines);
+    } else {
+        // Check that the buffer we have is not obsolete
+        if (tx->request_headers_raw_lines < list_size(tx->request_header_lines)) {
+            // Rebuild raw buffer
+            bstr_free(tx->request_headers_raw);
+            tx->request_headers_raw = htp_tx_generate_request_headers_raw(tx);
+            tx->request_headers_raw_lines = list_size(tx->request_header_lines);
+        }
+    }
+
+    return tx->request_headers_raw;
+}
